#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <LEDA/array.h>
#include <LEDA/basic.h>
#include <LEDA/color.h>
#include <LEDA/graph.h>
#include <LEDA/list.h>
#include <LEDA/map.h>
#include <LEDA/node_pq.h>
#include <LEDA/plane.h>
#include <LEDA/plane_alg.h>
#include <LEDA/sortseq.h>
#include <LEDA/stream.h>
#include <LEDA/ugraph.h>
#include <LEDA/window.h>

#define PI 3.14159265358979

window win( 1000, 750 );
color BG_color( white );
random_source random;
list<polygon> poly_list;

graph mbt;
node mbt_root;
map<node, point> mbt_cluster_center;
map<node, double> mbt_cluster_radius;
map<node, int> mbt_number_of_objects;
map<node, bool> mbt_and_mark;
map<node, bool> mbt_or_mark;
map<node, node> mbt_father;
map<node, node> mbt_left_son;
map<node, node> mbt_right_son;
map<node, bool> mbt_leaf;
struct mbt_object_item
{
 polygon poly;
 bool mark;
};
typedef list<mbt_object_item> mbt_object_list;
map<node, mbt_object_list> mbt_objects;
int mbt_truncation_size = 5;
int mbt_split_value_distance = 60;
bool mbt_display_construction = true;

UGRAPH<point, bool> vg;
list<polygon> vg_polygons;
map<node, polygon> vg_node_polygon;
map<node, vector> vg_wedge;
typedef list<node> sorted_adj_nodes;

double mp_dist;

point start;
point goal;
node start_node;
node goal_node;

bool exit_program = false;
bool file_has_changed = false;
bool objects_have_changed = true;
bool start_goal_is_fixed = false;

color marked_color( blue );
bool marked_filled = true;
color other_color( black );
bool other_filled = true;
color path_color( red );
line_style path_style = solid;
color start_goal_color( red );
int start_goal_radius = 3;
double ellipse_factor = 1.1;
int sm_insert = 1;
int output_single_query = 2;
int output_test_series = 2;
bool output_text = true;

struct ellipse
{
 point focus1;
 point focus2;
 double diameter;
};

enum query_type { em, sm, both };

int em_total_number;
int em_success_number;
array<int> em_object_count(41);
array<double> em_object_rate(41);
int sm_total_number;
int sm_success_number;
array<int> sm_object_count(41);
array<double> sm_object_rate(41);

list<segment> mp_path;
list<polygon> mp_objects;

int export_fill1 = 15;
int export_fill2 = 50;
int export_fill3 = 100;

string file_name = "untitled";
bool start_file = false;
string ync[3] = { "YES", "NO", "CANCEL" };

int film_index = 0;

void win_menu();
void win_menu_file();
void win_menu_file_new();
void win_menu_file_open();
void win_menu_file_open_execute();
void win_menu_file_read();
void win_menu_file_save();
void win_menu_file_saveas();
void win_menu_objects();
void win_menu_objects_insert();
void win_menu_objects_generate();
void win_menu_objects_delete();
void win_menu_objects_clear();
void win_menu_query();
void win_menu_query_test();
void win_menu_query_start_goal();
void win_menu_options();
void win_menu_options_mbt();
void win_menu_options_query();
void win_menu_options_display();
void win_menu_options_output();
void win_menu_options_export();
void win_menu_export();
void win_menu_exit();
void win_change();
void win_draw_polygons( bool );
void win_draw_polygons( list<polygon>&, bool );
void win_draw_polygons( list<polygon>&, color, bool, bool );
void win_draw_start();
void win_draw_start( color );
void win_draw_goal();
void win_draw_goal( color );

void file_read();
void file_save();
bool file_is_to_save();
void file_new_name();
bool file_get_name( bool, string );

void mbt_init();
void mbt_construction();
void mbt_construction( node, list<polygon>&, point, double, point );
point mbt_first_cluster_center();
double mbt_radius( point, list<polygon>&, point& );
bool mbt_is_root( node );
bool mbt_is_leaf( node );
void mbt_set_all_marks( bool );
void mbt_set_subtree_marks( node, bool );
list<polygon> mbt_get_tree( bool, bool, bool );
list<polygon> mbt_get_subtree( node, bool, bool, bool );
list<polygon> mbt_point_query( point, bool, bool, bool );
list<polygon> mbt_point_query( point, node, bool, bool, bool );
list<polygon> mbt_segment_query( segment, bool, bool, bool );
list<polygon> mbt_segment_query( segment, node, bool, bool, bool );
list<polygon> mbt_ellipse_query( ellipse, bool, bool, bool );
list<polygon> mbt_ellipse_query( ellipse, node, bool, bool, bool );
bool mbt_ellipse_query_prune_subtree( ellipse, node );
bool mbt_ellipse_query_accept_subtree( ellipse, node );

bool geom_segment_intersects_polygon( segment, polygon );
bool geom_segment_intersects_circle( segment, circle );
bool geom_ellipse_intersects_polygon( ellipse, polygon );
bool geom_ellipse_intersects_segment( ellipse, segment );
bool geom_point_in_ellipse( point, ellipse );
bool geom_path_in_ellipse( list<segment>&, ellipse );
bool geom_point_in_poly_list( point, list<polygon>& );
double geom_max_point_poly_distance( point, polygon, point& );
double geom_diameter();

edge vg_init();
bool vg_update( list<polygon>& );
bool vg_update( polygon );
bool vg_free( node );
bool vg_free( node, node );
bool vg_belongs_to_polygon( node, polygon );
node vg_insert_node( point p );
edge vg_insert_edge( node, node, bool );
bool vg_in_wedge( node, node );
list<node> vg_rotational_sweep( node, list<node>& );
bool vg_tangent( node, node );
bool vg_tangent( segment, polygon );
list<segment> vg_search();
void vg_draw_path( list<segment>& );
void vg_remove_path( list<segment>& );
void vg_draw_path( list<segment>&, color );
segment vg_segment( edge );
segment vg_segment( node, node );

void mp_test( query_type, int, int, int );
void mp_start( query_type, int );
void mp_ellipse_method( int );
void mp_segment_method( int );
void mp_show_statistics();
void mp_update_statistics( query_type, bool, double, int );
void mp_statistic_read();
void mp_statistic_write();
void mp_statistic_reset();
void mp_export_query();

polygon random_polygon( int, double );
point random_point( point, double, double );
point random_point( double, double, double, double );

color misc_new_color();
color misc_another_new_color();
color misc_no_path_color();
int misc_atoi( string );
void misc_export_to_eps();

void convert_to_postscript( list<point>&, list<segment>&, list<polygon>&,
                            list<polygon>&, list<polygon>&, int, int, int,
                            string, double, double, double, double );

main( int argc, char *argv[] )
{
 if ( argc > 1 )
 {
  file_name = argv[1];
  start_file = true;
 }
 win_menu();
}

void win_menu()
{
 int button_number;

 win.set_frame_label( "Hierarchical Motion Planning Using a Spatial Index" );
 win.set_redraw( win_change );

 win.button( "file" );
 win.button( "objects" );
 win.button( "query" );
 win.button( "options" );
 win.button( "export" );
 win.button( "EXIT" );
 win.buttons_per_line( 6 );

 win.display();
 win.init( 0, 1000, 0, 0 );
 win.clear();

 if ( start_file )
 {
  win_menu_file_open_execute();
  start_file = false;
 }

 while ( !exit_program )
 {
  button_number = win.read_mouse();
  switch( button_number )
  {
   case 0: win_menu_file();
           break;
   case 1: win_menu_objects();
           break;
   case 2: win_menu_query();
           break;
   case 3: win_menu_options();
           break;
   case 4: win_menu_export();
           break;
   case 5: win_menu_exit();
           break;
  }
 }
}

void win_menu_file()
{
 panel menu_file;
 menu_file.button( "new" );
 menu_file.button( "open" );
 menu_file.button( "read" );
 menu_file.button( "save" );
 menu_file.button( "save as" );
 menu_file.button( "CANCEL" );
 menu_file.buttons_per_line( 3 );

 int button_number = menu_file.open( win );

 switch ( button_number )
 {
  case 0: win_menu_file_new();
          break;
  case 1: win_menu_file_open();
          break;
  case 2: win_menu_file_read();
          break;
  case 3: win_menu_file_save();
          break;
  case 4: win_menu_file_saveas();
          break;
 }
}

void win_menu_file_new()
{
 if ( !file_has_changed || file_is_to_save() )
 {
  poly_list.clear();
  file_name = "untitled";
  file_new_name();
  objects_have_changed = true;
  win_draw_polygons( true );
 }
}

void win_menu_file_open()
{
 if ( !file_has_changed || file_is_to_save() )
  if ( file_get_name( true, "open" ) )
   win_menu_file_open_execute();
}

void win_menu_file_open_execute()
{
   poly_list.clear();
   file_read();
   win_draw_polygons( true );
   file_has_changed = false;
}

void win_menu_file_read()
{
 string old_name = file_name;
 if ( file_get_name( false, "read" ) )
 {
  list<polygon> old_polygons = poly_list;
  poly_list.clear();
  file_read();
  film_index = 0;
  win_draw_polygons( false );
  poly_list.conc( old_polygons );
  file_has_changed = true;
 }
 file_name = old_name;
}

void win_menu_file_save()
{
 if ( file_has_changed )
  if ( file_name == "untitled" )
   win_menu_file_saveas();
  else
   file_save();
}

void win_menu_file_saveas()
{
 if ( file_get_name( true, "save" ) )
  file_save();
}

void win_menu_objects()
{
 panel menu_objects;
 menu_objects.button( "insert" );
 menu_objects.button( "generate" );
 menu_objects.button( "delete" );
 menu_objects.button( "clear" );
 menu_objects.button( "redraw" );
 menu_objects.button( "CANCEL" );
 menu_objects.buttons_per_line( 3 );

 int button_number = menu_objects.open( win );
 switch ( button_number )
 {
  case 0: win_menu_objects_insert();
          break;
  case 1: win_menu_objects_generate();
          break;
  case 2: if ( poly_list.size() != 0 )
           win_menu_objects_delete();
          break;
  case 3: win_menu_objects_clear();
          break;
  case 4: win_draw_polygons( true );
          break;
 }
}

void win_menu_objects_insert()
{
 list<point> point_list;
 list<polygon> temp_list;
 point input;
 double x_input, y_input;
 window *window_pointer;

 int mouse_button = 0;

 panel menu_objects_insert;
 menu_objects_insert.button( "next" );
 menu_objects_insert.button( "quit" );
 menu_objects_insert.buttons_per_line( 2 );
 menu_objects_insert.display( win, 0, 0 );

 while ( mouse_button != 1 )
 {
  mouse_button = read_mouse( window_pointer, x_input, y_input );
  input = point( x_input, y_input );
  if ( ( window_pointer == &win ) && ( mouse_button >= 0 ) )
   mouse_button = 1;

  if ( mouse_button < 0 )
  {
   point_list.append( input );
   win.draw_point( input );
  }

  if ( ( mouse_button >= 0 ) && !point_list.empty() )
  {
   forall( input, point_list )
    win.draw_point( input, BG_color );
   if ( point_list.length() >= 3 )
   {
    poly_list.append( CONVEX_HULL( point_list ) );
    temp_list.clear();
    temp_list.append( poly_list.tail() );
    win_draw_polygons( temp_list, false );

    file_has_changed = true;
    objects_have_changed  = true;
   }
   point_list.clear();
  }

 }

}

void win_menu_objects_generate()
{
 int number_of_polygons = 100;
 int max_number_of_segments = 8;
 int max_size = 10;

 panel menu_objects_generate;
 menu_objects_generate.int_item( "number of polygons: ", number_of_polygons );
 menu_objects_generate.int_item( "max. number of segments: ",
				 max_number_of_segments, 3, 50 );
 menu_objects_generate.int_item( "max. size (% of diameter): ",
				 max_size, 1, 100 );
 menu_objects_generate.button( "OK" );
 menu_objects_generate.button( "CANCEL" );
 menu_objects_generate.buttons_per_line( 2 );

 int button_number = menu_objects_generate.open( win );
 if ( button_number == 0 && number_of_polygons > 0 )
 {
  list<polygon> temp_list;
  for ( int i = 1; i <= number_of_polygons; i++ )
   temp_list.append( random_polygon( max_number_of_segments,
                                   ( max_size / 100.0 ) * geom_diameter() ) );
  win_draw_polygons( temp_list, false );
  poly_list.conc( temp_list );

  file_has_changed = true;
  objects_have_changed = true;
 }
}

void win_menu_objects_delete()
{
 list<list_item> del_items;
 list<polygon> del_polys;
 list<polygon> undel_polys;
 list_item it;
 map<list_item, bool> del;
 point input;
 double x_input, y_input;
 window *window_pointer;

 int mouse_button = 0;

 panel menu_objects_delete;
 menu_objects_delete.button( "delete" );
 menu_objects_delete.button( "quit" );
 menu_objects_delete.buttons_per_line( 2 );
 menu_objects_delete.display( win, 0, 0 );

 while ( mouse_button != 1 )
 {
  mouse_button = read_mouse( window_pointer, x_input, y_input );
  input = point( x_input, y_input );
  if ( ( window_pointer == &win ) && ( mouse_button >= 0 ) )
   mouse_button = 1;
  if ( poly_list.empty() )
   mouse_button = 1;

  if ( mouse_button < 0 )
  {
   forall_items( it, poly_list )
   {
    if ( poly_list.contents( it ).inside( input ) )
     if ( del.defined( it ) )
      del[it] = !del[it];
     else
      del[it] = true;
   }
   del_items.clear();
   del_polys.clear();
   undel_polys.clear();
   forall_items( it, poly_list )
   {
    if ( del.defined( it ) )
     if ( del[it] )
     {
      del_items.append( it );
      del_polys.append( poly_list.contents( it ) );
     }
     else
      undel_polys.append( poly_list.contents( it ) );
   }
   win_draw_polygons( undel_polys, BG_color, true, false );
   win_draw_polygons( undel_polys, false );
   win_draw_polygons( del_polys, misc_new_color(), true, false );
  }

  if ( ( mouse_button == 0 ) && !del_items.empty() )
  {
   while ( !del_items.empty() )
    poly_list.del_item( del_items.pop() );
   file_has_changed = true;
   objects_have_changed = true;
   win_draw_polygons( true );
  }

 }

}

void win_menu_objects_clear()
{
 if ( !poly_list.empty() && win.confirm( "delete all objects?" ) )
 {
  poly_list.clear();
  file_has_changed = true;
  objects_have_changed = true;
  start_goal_is_fixed = false;
  mp_path.clear();
  win_draw_polygons( true );
 }
}

void win_menu_query()
{
 panel menu_query;
 menu_query.button( "ellipse method" );
 menu_query.button( "segment method" );
 menu_query.button( "test series" );
 menu_query.button( "start/goal" );
 menu_query.button( "statistics" );
 menu_query.button( "CANCEL" );
 menu_query.buttons_per_line( 3 );

 int button_number = menu_query.open( win );
 switch ( button_number )
 {
  case 0: mp_start( em, output_single_query );
          break;
  case 1: mp_start( sm, output_single_query );
          break;
  case 2: win_menu_query_test();
          break;
  case 3: win_menu_query_start_goal();
          break;
  case 4: mp_show_statistics();
          break;
 }

}

void win_menu_query_test()
{
 int min_start_goal_distance = 0;
 int max_start_goal_distance = 60;
 int number_of_passes = 10;

 panel menu_query_test;
 menu_query_test.int_item( "min. distance start/goal (%): ",
                           min_start_goal_distance, 0, 100 );
 menu_query_test.int_item( "max. distance start/goal (%): ",
                           max_start_goal_distance, 0, 100 );
// menu_query_test.int_item( "number of passes: ",
//                           number_of_passes, 0, 100 );
 menu_query_test.int_item( "number of passes: ",
                           number_of_passes );
 menu_query_test.button( "ellipse method" );
 menu_query_test.button( "segment method" );
 menu_query_test.button( "both methods" );
 menu_query_test.button( "CANCEL" );
 menu_query_test.buttons_per_line( 2 );

 int button_number = menu_query_test.open( win );
 if ( button_number < 3 &&
      min_start_goal_distance > max_start_goal_distance )
  win_menu_query_test();
 else
  switch ( button_number )
  {
   case 0: mp_test( em, min_start_goal_distance,
                    max_start_goal_distance, number_of_passes );
           break;
   case 1: mp_test( sm, min_start_goal_distance,
                    max_start_goal_distance, number_of_passes );
           break;
   case 2: mp_test( both, min_start_goal_distance,
                    max_start_goal_distance, number_of_passes );
           break;
  }
}

void win_menu_query_start_goal()
{
 point input;
 double x_input, y_input;
 window *window_pointer;
 int mode = 2;
 bool start_is_fixed = start_goal_is_fixed;
 bool goal_is_fixed = start_goal_is_fixed;
 bool redraw = false;

 int mouse_button = 0;

 panel menu_query_start_goal;
 menu_query_start_goal.button( "start" );
 menu_query_start_goal.button( "goal" );
 menu_query_start_goal.button( "quit" );
 menu_query_start_goal.buttons_per_line( 3 );
 menu_query_start_goal.display( win, 0, 0 );

 while ( mouse_button != 2 )
 {
  mouse_button = read_mouse( window_pointer, x_input, y_input );
  input = point( x_input, y_input );
  if ( ( window_pointer == &win ) && ( mouse_button >= 0 ) )
   mouse_button = 2;
  switch ( mouse_button )
  {
   case 0: start_is_fixed = false;
           redraw = true;
           mode = 0;
           vg_remove_path( mp_path );
           win_draw_start( BG_color );
           mp_path.clear();
           break;
   case 1: goal_is_fixed = false;
           redraw = true;
           mode = 1;
           vg_remove_path( mp_path );
           win_draw_goal( BG_color );
           mp_path.clear();
           break;
   case 2: start_goal_is_fixed = start_is_fixed && goal_is_fixed;
           win_draw_polygons( redraw );
           break;
  }
  if ( mouse_button < 0 )
   switch ( mode )
   {
    case 0: start = input;
            mode = 2;
            start_is_fixed = true;
            win_draw_start();
            break;
    case 1: goal = input;
            mode = 2;
            goal_is_fixed = true;
            win_draw_goal();
            break;
   }

 }
}

void win_menu_options()
{
 panel menu_options;
 menu_options.button( "mbt" );
 menu_options.button( "query" );
 menu_options.button( "display" );
 menu_options.button( "output" );
 menu_options.button( "export" );
 menu_options.button( "CANCEL" );
 menu_options.buttons_per_line( 3 );

 int button_number = menu_options.open( win );
 switch ( button_number )
 {
  case 0: win_menu_options_mbt();
          break;
  case 1: win_menu_options_query();
          break;
  case 2: win_menu_options_display();
          break;
  case 3: win_menu_options_output();
          break;
  case 4: win_menu_options_export();
          break;
 }
}

void win_menu_options_mbt()
{
 int old_truncation_size = mbt_truncation_size;
 int old_split_value_distance = mbt_split_value_distance;
 int old_mbt_display_construction = mbt_display_construction;

 panel menu_options_mbt;
 menu_options_mbt.int_item( "truncation size: ", mbt_truncation_size, 1, 50 );
 menu_options_mbt.int_item( "distance old/new split value (%): ",
			    mbt_split_value_distance, 0, 100 );
 menu_options_mbt.bool_item( "display mbt construction: ",
			     mbt_display_construction );
 menu_options_mbt.button( "OK" );
 menu_options_mbt.button( "CANCEL" );

 int button_number = menu_options_mbt.open( win );
 switch ( button_number )
 {
  case 0: objects_have_changed = true;
          break;
  case 1: mbt_truncation_size = old_truncation_size;
          mbt_split_value_distance = old_split_value_distance;
          mbt_display_construction = old_mbt_display_construction;
          break;
 }
}

void win_menu_options_query()
{
 double old_ellipse_factor = ellipse_factor;
 int old_sm_insert = sm_insert;

 panel menu_options_query;
 menu_options_query.real_item( "ellipse growing by factor: ",
                                 ellipse_factor );
 menu_options_query.choice_item( "segment method: insert obstacles: ",
                                 sm_insert, "first", "all" );
 menu_options_query.button( "OK" );
 menu_options_query.button( "CANCEL" );

 int button_number = menu_options_query.open( win );
 if ( button_number == 1 )
 {
  ellipse_factor = old_ellipse_factor;
  sm_insert = old_sm_insert;
 }

}

void win_menu_options_display()
{
 color old_marked_color = marked_color;
 bool old_marked_filled = marked_filled;
 color old_other_color = other_color;
 bool old_other_filled = other_filled;
 color old_path_color = path_color;
 line_style old_path_style = path_style;
 color old_start_goal_color = start_goal_color;
 int old_start_goal_radius = start_goal_radius;

 panel menu_options_display;
 menu_options_display.color_item( "marked polygons color: ", marked_color );
 menu_options_display.bool_item( "marked polygons fill: ", marked_filled );
 menu_options_display.color_item( "other polygons color: ", other_color);
 menu_options_display.bool_item( "other polygons fill: ", other_filled );
 menu_options_display.color_item( "path color: ", path_color );
 menu_options_display.lstyle_item( "path line style: ", path_style );
 menu_options_display.color_item( "start/goal color: ", start_goal_color );
 menu_options_display.int_item( "start/goal radius: ",
			        start_goal_radius, 1, 20 );
 menu_options_display.button( "OK" );
 menu_options_display.button( "CANCEL" );
 menu_options_display.buttons_per_line( 2 );

 int button_number = menu_options_display.open( win );
 switch ( button_number )
 {
  case 0: win_draw_polygons( true );
          break;
  case 1: marked_color = old_marked_color;
          marked_filled = old_marked_filled;
          other_color = old_other_color;
          other_filled = old_other_filled;
          path_color = old_path_color;
          path_style = old_path_style;
          start_goal_color = old_start_goal_color;
          start_goal_radius = old_start_goal_radius;
          break;
 }
}

void win_menu_options_output()
{
 int old_output_single_query = output_single_query;
 int old_output_test_series = output_test_series;
 bool old_output_text = output_text;

 panel menu_options_output;
 menu_options_output.choice_item( "single query: " , output_single_query,
                                  "none", "result", "run", "step" );
 menu_options_output.choice_item( "test series: " , output_test_series,
                                  "none", "result", "run", "step" );
 menu_options_output.bool_item( "text: ", output_text );
 menu_options_output.button( "OK" );
 menu_options_output.button( "CANCEL" );
 menu_options_output.buttons_per_line( 2 );

 int button_number = menu_options_output.open( win );
 switch ( button_number )
 {
  case 1: output_single_query = old_output_single_query;
          output_test_series = old_output_test_series;
          output_text = old_output_text;
          break;
 }
}

void win_menu_options_export()
{
 int old_export_fill1 = export_fill1;
 int old_export_fill2 = export_fill2;
 int old_export_fill3 = export_fill3;

 panel menu_options_export;
 menu_options_export.text_item( "export objects: fill intensity" );
 menu_options_export.int_item( "list 1: ", export_fill1, 0, 100 );
 menu_options_export.int_item( "list 2: ", export_fill2, 0, 100 );
 menu_options_export.int_item( "list 3: ", export_fill3, 0, 100 );
 menu_options_export.button( "OK" );
 menu_options_export.button( "CANCEL" );
 menu_options_export.buttons_per_line( 2 );

 int button_number = menu_options_export.open( win );
 if ( button_number == 1 )
 {
  export_fill1 = old_export_fill1;
  export_fill2 = old_export_fill2;
  export_fill3 = old_export_fill3;
 }
}

void win_menu_export()
{
 misc_export_to_eps();
}

void win_menu_exit()
{
 if ( !file_has_changed || file_is_to_save() )
  exit_program = true;
}

void win_change()
{
 win_draw_polygons( true );
}

void win_draw_polygons( bool clear )
{
 if ( objects_have_changed )
  win_draw_polygons( poly_list, clear );
 else
 {
  win_draw_polygons( mbt_get_subtree( mbt_root, false, true, false ),
                     other_color, other_filled, clear );
  win_draw_polygons( mbt_get_subtree( mbt_root, true, false, false ),
                     marked_color, marked_filled, false );
 }
}

void win_draw_polygons( list<polygon>& pol_list, bool clear )
{
 win_draw_polygons( pol_list, other_color, other_filled, clear );
}

void win_draw_polygons( list<polygon>& pol_list, color pol_color,
                        bool filled, bool clear )
{
 polygon pol;
 if ( clear )
  win.clear();
 if ( !pol_list.empty() )
 {
  forall(pol, pol_list)
  {
   if ( filled )
    win.draw_filled_polygon( pol, pol_color );
   else
    win.draw_polygon( pol, pol_color );
  }
 }
 if ( clear )
  vg_draw_path( mp_path );
 if ( start_goal_is_fixed )
 {
  win_draw_start();
  win_draw_goal();
 }
}

void win_draw_start()
{
 win_draw_start( start_goal_color );
}

void win_draw_start( color start_color )
{
 win.draw_disc( start, start_goal_radius, start_color );
}

void win_draw_goal()
{
 win_draw_goal( start_goal_color );
}

void win_draw_goal( color goal_color )
{
 win.draw_disc( goal, start_goal_radius, goal_color );
}

void file_read()
{
 if ( output_text )
  cout << "reading file '" << file_name << "'..." << endl;
 file_istream from_file( file_name );
 poly_list.read( from_file, EOF );
 objects_have_changed = true;
 if ( output_text )
  cout << "...done" << endl << endl;
}

void file_save()
{
 if ( output_text )
  cout << "saving file '" << file_name << "'..." << endl;
 file_ostream into_file( file_name );
 poly_list.print( into_file );
 file_has_changed = false;
 if ( output_text )
  cout << "...done" << endl << endl;
}

bool file_is_to_save()
{
 int result = win.read_panel( "'" + file_name +
                              "' has been modified. Save?", 3, ync );
 if ( result == 0 )
 {
  if ( file_name == "untitled" && !file_get_name( true, "save" ) )
   return false;
  else
   file_save();
 }
 return ( result != 2 );
}

void file_new_name()
{
 string label = "Hierarchical Motion Planning Using a Spatial Index";
 if ( file_name != "untitled" )
  label += ( " --- " + file_name );
 win.set_frame_label( label );
}

bool file_get_name( bool change_name, string action )
{
 string new_name = file_name;
 panel get_name;
 get_name.string_item( "file to " + action + ": ", new_name );
 get_name.button( "OK" );
 get_name.button( "CANCEL" );
 if ( get_name.open( win ) == 0 )
 {
  file_name = new_name;
  if ( change_name )
   file_new_name();
  return true;
 }
 else
  return false;
}

void mbt_init()
{
 mbt.clear();
 mbt_root = mbt.new_node();
}

void mbt_construction()
{
 mbt_init();
 point extreme;
 if ( output_text )
  cout << "constructing mbt for " << poly_list.size() << " objects..." << endl;
 point first_cluster_center = mbt_first_cluster_center();
 double first_cluster_radius = mbt_radius( first_cluster_center,
					   poly_list, extreme );
 mbt_construction( mbt_root, poly_list, first_cluster_center,
	           first_cluster_radius, extreme );
 mbt_set_all_marks( false );
 poly_list = mbt_get_tree( true, true, false );
 if ( output_text )
  cout << "...done. mbt consists of " << mbt.number_of_nodes()
       << " nodes." << endl << endl;
 objects_have_changed = false;
}

void mbt_construction( node current_node, list<polygon>& current_pol_list,
                       point cluster_center1, double cluster_radius1,
                       point extremal_point1 )
{
 mbt_cluster_center[current_node] = cluster_center1;
 mbt_cluster_radius[current_node] = cluster_radius1;
 mbt_number_of_objects[current_node] = current_pol_list.size();

 if ( current_pol_list.size() > mbt_truncation_size )
 {
  double lambda = mbt_split_value_distance / 100.0;
  point cluster_center2( ( 1 - lambda ) * cluster_center1.xcoord() +
                                 lambda * extremal_point1.xcoord(),
                         ( 1 - lambda ) * cluster_center1.ycoord() +
                                 lambda * extremal_point1.ycoord() );
  cluster_radius1 = 0;
  double cluster_radius2 = 0;
  double distance1;
  double distance2;
  point extremal_point2;
  point p1;
  point p2;
  polygon pol;
  list<polygon> pol_list1;
  list<polygon> pol_list2;

  while ( !current_pol_list.empty() )
  {
   pol = current_pol_list.pop();
   distance1 = geom_max_point_poly_distance( cluster_center1, pol, p1 );
   distance2 = geom_max_point_poly_distance( cluster_center2, pol, p2 );
   if ( distance1 < distance2 )
   {
    pol_list1.append( pol );
    if ( distance1 > cluster_radius1 )
    {
     cluster_radius1 = distance1;
     extremal_point1 = p1;
    }
   }
   else
   {
    pol_list2.append( pol );
    if ( distance2 > cluster_radius2 )
    {
     cluster_radius2 = distance2;
     extremal_point2 = p2;
    }
   }
  }

  if ( pol_list1.empty() )
  {
   cluster_center1 = cluster_center2;
   while ( pol_list1.size() < pol_list2.size() )
    pol_list1.append( pol_list2.pop() );
   cluster_radius1 = mbt_radius( cluster_center1, pol_list1, extremal_point1 );
   cluster_radius2 = mbt_radius( cluster_center2, pol_list2, extremal_point2 );
  }
  if ( pol_list2.empty() )
  {
   cluster_center2 = cluster_center1;
   while ( pol_list2.size() < pol_list1.size() )
    pol_list2.append( pol_list1.pop() );
   cluster_radius1 = mbt_radius( cluster_center1, pol_list1, extremal_point1 );
   cluster_radius2 = mbt_radius( cluster_center2, pol_list2, extremal_point2 );
  }

  if ( mbt_display_construction )
  {
   win_draw_polygons( pol_list1, misc_new_color(), true, false );
   win_draw_polygons( pol_list2, misc_another_new_color(), true, false );
  }

  node left_son = mbt.new_node();
  node right_son = mbt.new_node();
  mbt_father[left_son] = current_node;
  mbt_father[right_son] = current_node;
  mbt_left_son[current_node] = left_son;
  mbt_right_son[current_node] = right_son;
  mbt_leaf[current_node] = false;

  mbt_construction( left_son, pol_list1, cluster_center1,
		    cluster_radius1, extremal_point1 );
  mbt_construction( right_son, pol_list2, cluster_center2,
		    cluster_radius2, extremal_point2 );
 }

 else
 {
  mbt_object_item object_item;
  mbt_object_list object_list;

  while( !current_pol_list.empty() )
  {
   object_item.poly = current_pol_list.pop();
   object_item.mark = false;
   object_list.append( object_item );
  }
  mbt_objects[current_node] = object_list;
  mbt_leaf[current_node] = true;

  if ( mbt_display_construction )
   forall( object_item, object_list )
    win.draw_filled_polygon( object_item.poly, other_color );
 }
}

point mbt_first_cluster_center()
{
 return point( ( 2*win.xmin() + win.xmax() ) / 3,
               ( 2*win.ymin() + win.ymax() ) / 3 );
}

double mbt_radius( point cluster_center, list<polygon>& pol_list,
		   point& extreme)
{
 point p;
 double radius = 0;
 double distance;
 polygon pol;
 if ( !pol_list.empty() )
  forall( pol, pol_list )
  {
   distance = geom_max_point_poly_distance( cluster_center, pol, p );
   if ( distance > radius )
   {
    radius = distance;
    extreme = p;
   }
  }
 return radius;
}

bool mbt_is_root( node current_node )
{
 return ( current_node == mbt_root );
}

bool mbt_is_leaf( node current_node )
{
 return ( mbt_leaf.defined( current_node ) && mbt_leaf[current_node] );
}

void mbt_set_all_marks( bool new_mark )
{
 mbt_set_subtree_marks( mbt_root, new_mark );
}

void mbt_set_subtree_marks( node current_node, bool new_mark )
{
 mbt_and_mark[current_node] = new_mark;
 mbt_or_mark[current_node] = new_mark;

 if ( mbt_is_leaf( current_node ) )
 {
  mbt_object_item object_item;
  mbt_object_list object_list = mbt_objects[current_node];
  mbt_object_list new_object_list;
  while( !object_list.empty() )
  {
   object_item = object_list.pop();
   object_item.mark = new_mark;
   new_object_list.append( object_item );
  }
  mbt_objects[current_node] = new_object_list;
 }

 else
 {
  mbt_set_subtree_marks( mbt_left_son[current_node], new_mark );
  mbt_set_subtree_marks( mbt_right_son[current_node], new_mark );
 }
}

list<polygon> mbt_get_tree( bool get_marked_objects,
                            bool get_unmarked_objects,
                            bool mark_new_objects )
{
 return mbt_get_subtree( mbt_root, get_marked_objects,
			 get_unmarked_objects, mark_new_objects );
}

list<polygon> mbt_get_subtree( node current_node, bool get_marked_objects,
                               bool get_unmarked_objects,
                               bool mark_new_objects )
{
 list<polygon> result;
 list<polygon> result_left_son;
 list<polygon> result_right_son;

 if ( mbt_is_leaf( current_node ) )
 {
  mbt_object_item object_item;
  mbt_object_list object_list = mbt_objects[current_node];
  mbt_object_list new_object_list;
  bool and_mark = true;
  bool or_mark = false;
  while( !object_list.empty() )
  {
   object_item = object_list.pop();
   if ( ( get_marked_objects && object_item.mark ) ||
        ( get_unmarked_objects && !object_item.mark ) )
   {
    result.append( object_item.poly );
    if ( mark_new_objects )
     object_item.mark = true;
   }
   and_mark = and_mark && object_item.mark;
   or_mark = or_mark || object_item.mark;
   new_object_list.append( object_item );
  }
  mbt_objects[current_node] = new_object_list;
  if ( mark_new_objects )
  {
   mbt_and_mark[current_node] = and_mark;
   mbt_or_mark[current_node] = or_mark;
  }
 }

 else
  if ( ( get_marked_objects && mbt_or_mark[current_node] ) ||
       ( get_unmarked_objects && !mbt_and_mark[current_node] ) )
  {
   result_left_son = mbt_get_subtree( mbt_left_son[current_node],
                                      get_marked_objects,
                                      get_unmarked_objects,
                                      mark_new_objects ) ;
   result_right_son = mbt_get_subtree( mbt_right_son[current_node],
                                       get_marked_objects,
                                       get_unmarked_objects,
                                       mark_new_objects ) ;
   result.conc( result_left_son );
   result.conc( result_right_son );
   if ( mark_new_objects )
   {
    mbt_and_mark[current_node] = mbt_and_mark[mbt_left_son[current_node]] &&
                                 mbt_and_mark[mbt_right_son[current_node]];
    mbt_or_mark[current_node] = mbt_or_mark[mbt_left_son[current_node]] ||
                                mbt_or_mark[mbt_right_son[current_node]];
   }
  }

 return result;
}

list<polygon> mbt_point_query( point p, bool get_marked_objects,
                               bool get_unmarked_objects,
                               bool mark_new_objects )
{
 return mbt_point_query( p, mbt_root, get_marked_objects,
                         get_unmarked_objects, mark_new_objects );
}

list<polygon> mbt_point_query( point p, node current_node,
                               bool get_marked_objects,
                               bool get_unmarked_objects,
                               bool mark_new_objects )
{
 list<polygon> result;
 list<polygon> result_left_son;
 list<polygon> result_right_son;

 if ( mbt_is_leaf( current_node ) )
 {
  mbt_object_item object_item;
  mbt_object_list object_list = mbt_objects[current_node];
  mbt_object_list new_object_list;
  bool and_mark = true;
  bool or_mark = false;
  while( !object_list.empty() )
  {
   object_item = object_list.pop();
   if ( ( ( get_marked_objects && object_item.mark ) ||
          ( get_unmarked_objects && !object_item.mark ) )
        && object_item.poly.inside( p ) )
   {
    result.append( object_item.poly );
    if ( mark_new_objects )
     object_item.mark = true;
   }
   and_mark = and_mark && object_item.mark;
   or_mark = or_mark || object_item.mark;
   new_object_list.append( object_item );
  }
  mbt_objects[current_node] = new_object_list;
  if ( mark_new_objects )
  {
   mbt_and_mark[current_node] = and_mark;
   mbt_or_mark[current_node] = or_mark;
  }
 }

 else
  if ( ( ( get_marked_objects && mbt_or_mark[current_node] ) ||
       ( get_unmarked_objects && !mbt_and_mark[current_node] ) )
       && circle( mbt_cluster_center[current_node],
                  mbt_cluster_radius[current_node] ).inside( p ) )
  {
   result_left_son = mbt_point_query( p, mbt_left_son[current_node],
                                      get_marked_objects,
                                      get_unmarked_objects,
                                      mark_new_objects ) ;
   result_right_son = mbt_point_query( p, mbt_right_son[current_node],
                                       get_marked_objects,
                                       get_unmarked_objects,
                                       mark_new_objects ) ;
   result.conc( result_left_son );
   result.conc( result_right_son );
   if ( mark_new_objects )
   {
    mbt_and_mark[current_node] = mbt_and_mark[mbt_left_son[current_node]] &&
                                 mbt_and_mark[mbt_right_son[current_node]];
    mbt_or_mark[current_node] = mbt_or_mark[mbt_left_son[current_node]] ||
                                mbt_or_mark[mbt_right_son[current_node]];
   }
  }

 return result;
}

list<polygon> mbt_segment_query( segment s,
                                 bool get_marked_objects,
                                 bool get_unmarked_objects,
                                 bool mark_new_objects )
{
 return mbt_segment_query( s, mbt_root, get_marked_objects,
                           get_unmarked_objects, mark_new_objects );
}

list<polygon> mbt_segment_query( segment s, node current_node,
                                 bool get_marked_objects,
                                 bool get_unmarked_objects,
                                 bool mark_new_objects )
{
 list<polygon> result;
 list<polygon> result_left_son;
 list<polygon> result_right_son;
 circle cluster( mbt_cluster_center[current_node],
                 mbt_cluster_radius[current_node] );

 if ( mbt_is_leaf( current_node ) )
 {
  mbt_object_item object_item;
  mbt_object_list object_list = mbt_objects[current_node];
  mbt_object_list new_object_list;
  bool and_mark = true;
  bool or_mark = false;
  while( !object_list.empty() )
  {
   object_item = object_list.pop();
   if ( ( ( get_marked_objects && object_item.mark ) ||
          ( get_unmarked_objects && !object_item.mark ) )
        && geom_segment_intersects_polygon( s, object_item.poly ) )
   {
    result.append( object_item.poly );
    if ( mark_new_objects )
     object_item.mark = true;
   }
   and_mark = and_mark && object_item.mark;
   or_mark = or_mark || object_item.mark;
   new_object_list.append( object_item );
  }
  mbt_objects[current_node] = new_object_list;
  if ( mark_new_objects )
  {
   mbt_and_mark[current_node] = and_mark;
   mbt_or_mark[current_node] = or_mark;
  }
 }

 else
  if ( ( ( get_marked_objects && mbt_or_mark[current_node] ) ||
         ( get_unmarked_objects && !mbt_and_mark[current_node] ) )
       && geom_segment_intersects_circle( s, cluster ) )
  {
   result_left_son = mbt_segment_query( s, mbt_left_son[current_node],
                                        get_marked_objects,
                                        get_unmarked_objects,
                                        mark_new_objects ) ;
   result_right_son = mbt_segment_query( s, mbt_right_son[current_node],
                                         get_marked_objects,
                                         get_unmarked_objects,
                                         mark_new_objects ) ;
   result.conc( result_left_son );
   result.conc( result_right_son );
   if ( mark_new_objects )
   {
    mbt_and_mark[current_node] = mbt_and_mark[mbt_left_son[current_node]] &&
                                 mbt_and_mark[mbt_right_son[current_node]];
    mbt_or_mark[current_node] = mbt_or_mark[mbt_left_son[current_node]] ||
                                mbt_or_mark[mbt_right_son[current_node]];
   }
  }

 return result;
}

list<polygon> mbt_ellipse_query( ellipse e,
                                 bool get_marked_objects,
                                 bool get_unmarked_objects,
                                 bool mark_new_objects )
{
 return mbt_ellipse_query( e, mbt_root, get_marked_objects,
                           get_unmarked_objects, mark_new_objects );
}

list<polygon> mbt_ellipse_query( ellipse e, node current_node,
                                 bool get_marked_objects,
                                 bool get_unmarked_objects,
                                 bool mark_new_objects )
{
 list<polygon> result;
 list<polygon> result_left_son;
 list<polygon> result_right_son;

 if ( mbt_is_leaf( current_node ) )
 {
  mbt_object_item object_item;
  mbt_object_list object_list = mbt_objects[current_node];
  mbt_object_list new_object_list;
  bool and_mark = true;
  bool or_mark = false;
  while( !object_list.empty() )
  {
   object_item = object_list.pop();
   if ( ( ( get_marked_objects && object_item.mark ) ||
          ( get_unmarked_objects && !object_item.mark ) )
        && geom_ellipse_intersects_polygon( e, object_item.poly ) )
   {
    result.append( object_item.poly );
    if ( mark_new_objects )
     object_item.mark = true;
   }
   and_mark = and_mark && object_item.mark;
   or_mark = or_mark || object_item.mark;
   new_object_list.append( object_item );
  }
  mbt_objects[current_node] = new_object_list;
  if ( mark_new_objects )
  {
   mbt_and_mark[current_node] = and_mark;
   mbt_or_mark[current_node] = or_mark;
  }
 }

 else
  if ( ( ( get_marked_objects && mbt_or_mark[current_node] ) ||
         ( get_unmarked_objects && !mbt_and_mark[current_node] ) )
       && !mbt_ellipse_query_prune_subtree( e, current_node ) )
  {
   if ( mbt_ellipse_query_accept_subtree( e, current_node ) )
    result.conc( mbt_get_subtree( current_node, get_marked_objects,
                                  get_unmarked_objects, mark_new_objects ) );
   else
   {
    result_left_son = mbt_ellipse_query( e, mbt_left_son[current_node],
                                         get_marked_objects,
                                         get_unmarked_objects,
                                         mark_new_objects ) ;
    result_right_son = mbt_ellipse_query( e, mbt_right_son[current_node],
                                          get_marked_objects,
                                          get_unmarked_objects,
                                          mark_new_objects ) ;
    result.conc( result_left_son );
    result.conc( result_right_son );
   }
   if ( mark_new_objects )
   {
    mbt_and_mark[current_node] = mbt_and_mark[mbt_left_son[current_node]] &&
                                 mbt_and_mark[mbt_right_son[current_node]];
    mbt_or_mark[current_node] = mbt_or_mark[mbt_left_son[current_node]] ||
                                mbt_or_mark[mbt_right_son[current_node]];
   }
  }

 return result;
}

bool mbt_ellipse_query_prune_subtree( ellipse e, node current_node )
{
 return ( mbt_cluster_center[current_node].distance( e.focus1 ) +
          mbt_cluster_center[current_node].distance( e.focus2 ) -
          2 * mbt_cluster_radius[current_node] >= e.diameter );
}

bool mbt_ellipse_query_accept_subtree( ellipse e, node current_node )
{
 return ( mbt_cluster_center[current_node].distance( e.focus1 ) +
          mbt_cluster_center[current_node].distance( e.focus2 ) +
          2 * mbt_cluster_radius[current_node] <= e.diameter );
}

bool geom_segment_intersects_polygon( segment s, polygon pol )
{
 return ( pol.inside( s.source() ) || pol.inside( s.target() ) ||
	  !pol.intersection( s ).empty() );
}

bool geom_segment_intersects_circle( segment s, circle c )
{
 if ( c.inside( s.source() ) || c.inside( s.target() ) )
  return true;
 else
 {
  double sx = s.dx();
  double sy = s.dy();
  double cx = s.xcoord1() - c.center().xcoord();
  double cy = s.ycoord1() - c.center().ycoord();
  double lambda = -( sx*cx + sy*cy )/( sx*sx + sy*sy );
  if ( lambda > 0 && lambda < 1 )
  {
   point p( s.xcoord1() + lambda*sx, s.ycoord1() + lambda*sy );
   return c.inside( p );
  }
  else
   return false;
 }
}

bool geom_ellipse_intersects_polygon( ellipse e, polygon pol )
{
 segment s;
 bool intersects = ( pol.inside( e.focus1 ) || pol.inside( e.focus2 ) );
 if ( !intersects )
  forall( s, pol.segments() )
   intersects = intersects || geom_ellipse_intersects_segment( e, s );
 return intersects;
}

bool geom_ellipse_intersects_segment( ellipse e, segment s )
{
 vector translation( ( -e.focus1.xcoord() - e.focus2.xcoord() ) / 2,
		     ( -e.focus1.ycoord() - e.focus2.ycoord() ) / 2);
 double rotation = -segment( e.focus1, e.focus2 ).angle();
 double a = e.diameter / 2;
 double c = e.focus1.distance( e.focus2 ) / 2;
 double a_square = a*a;
 double b_square = a_square - c*c;

 point inter;
 if ( a <= c )
  return s.intersection( segment( e.focus1, e.focus2 ), inter );

 if ( geom_point_in_ellipse( s.source(), e ) ||
      geom_point_in_ellipse( s.target(), e ) )
  return true;
 else

  if ( s.length() == 0 )
   return geom_point_in_ellipse( s.source(), e );
  else
  {
   s = s.translate( translation ).rotate( point( 0, 0 ), rotation );
   double coeff1 =  s.dx() * s.dx() / a_square +
                    s.dy() * s.dy() / b_square;
   double coeff2 =  s.xcoord1() * s.dx() / a_square +
                    s.ycoord1() * s.dy() / b_square;
   double coeff3 =  s.xcoord1() * s.xcoord1() / a_square +
                    s.ycoord1() * s.ycoord1() / b_square - 1;
   double discrim = coeff2 * coeff2 - coeff1 * coeff3;
   if ( discrim < 0 )
    return false;
   else
   {
    double sol1 = ( -coeff2 + sqrt( discrim ) ) / coeff1;
    double sol2 = ( -coeff2 - sqrt( discrim ) ) / coeff1;
    if ( ( sol1 < 0 || sol1 > 1 ) && ( sol2 < 0 || sol2 > 1 ) )
     return false;
    else
     return true;
   }
  }
}

bool geom_point_in_ellipse( point p, ellipse e )
{
 return ( p.distance( e.focus1 ) + p.distance( e.focus2 ) <= e.diameter );
}

bool geom_path_in_ellipse( list<segment>& path, ellipse e )
{
 segment s;

 if ( geom_point_in_ellipse( path.tail().target(), e ) )
 {
  forall( s, path )
   if ( !geom_point_in_ellipse( s.source(), e  ) )
    return false;
 }
 else
  return false;

 return true;
}

bool geom_point_in_poly_list( point p, list<polygon>& pol_list )
{
 polygon pol;
 forall( pol, pol_list )
  if ( pol.inside( p ) )
   return true;
 return false;
}

double geom_max_point_poly_distance( point p1, polygon pol, point& extreme )
{
 point p2;
 double max = 0;
 double distance;
 forall( p2, pol.vertices() )
 {
  distance = p1.distance( p2 );
  if ( distance > max )
  {
   max = distance;
   extreme = p2;
  }
 }
 return max;
}

double geom_diameter()
{
 double x = win.xmax() - win.xmin();
 double y = win.ymax() - win.ymin();
 return sqrt( x*x + y*y );
}

edge vg_init()
{
 if ( output_text )
  cout << "initializing vg" << endl << endl;
 vg_polygons.clear();
 vg.clear();
 return vg.new_edge( start_node = vg_insert_node( start ),
                     goal_node = vg_insert_node( goal ), true );
}

bool vg_update( list<polygon>& pol_list )
{
 if ( !pol_list.empty() )
 {
  polygon pol;
  if ( output_text )
   cout << "updating vg from " << vg_polygons.size() << " to " <<
           vg_polygons.size() + pol_list.size() << " objects..." << endl;
  forall( pol, pol_list )
   if ( !vg_update( pol ) )
    return false;
  if ( output_text )
   cout << "...done. vg consists of "  << vg.number_of_nodes() <<
           " nodes and " << vg.number_of_edges() << " edges." << endl << endl;
 }
 return true;
}

bool vg_update( polygon pol )
{
 if ( pol.inside( start ) || pol.inside( goal ) )
  return false;

 node current_node;
 edge current_edge;
 list_item current_item;

 // step 1

 list<node> non_free_nodes;
 forall_nodes( current_node, vg )
 if ( pol.inside( vg.inf( current_node ) ) )
   non_free_nodes.append( current_node );
 while ( !non_free_nodes.empty() )
  vg.del_node( non_free_nodes.pop() );

 // step 2

 list<node> new_nodes;
 list<node> old_nodes = vg.all_nodes();
 list<edge> old_edges = vg.all_edges();
 map<node, bool> vg_nodes_free;
 point p;

 forall( p, pol.vertices() )
 {
  current_node = vg_insert_node( p );
  new_nodes.append( current_node );
  vg_node_polygon[current_node] = pol;
  vg_nodes_free[current_node] = vg_free( current_node );
 }

 node pred_node;
 node succ_node;
 list<list_item> non_free_items;

 forall_items( current_item, new_nodes )
 {
  current_node = new_nodes.inf( current_item );
  pred_node = new_nodes.inf( new_nodes.cyclic_pred( current_item ) );
  succ_node = new_nodes.inf( new_nodes.cyclic_succ( current_item ) );
  if ( vg_nodes_free[current_node] &&  vg_nodes_free[succ_node] &&
       vg_free( current_node, succ_node ) )
  vg_insert_edge( current_node, succ_node, true );
  vg_wedge[current_node] =
   vector( segment( vg.inf( current_node ), vg.inf( pred_node ) ).angle(),
           segment( vg.inf( current_node ), vg.inf( succ_node ) ).angle() );
  if ( !vg_nodes_free[current_node] )
   non_free_items.append( current_item );
 }

 while ( !non_free_items.empty() )
  vg.del_node( new_nodes.del_item( non_free_items.pop() ) );

 // step 3

 node old_node;
 list<node> free_nodes;

 forall( current_node, new_nodes )
 {
  free_nodes = vg_rotational_sweep( current_node, old_nodes );
  forall( old_node, free_nodes )
  {
   if ( vg_tangent( current_node, old_node ) )
    vg_insert_edge( current_node, old_node, true );
   else
    vg_insert_edge( current_node, old_node, false );
  }
 }

 // step 4

 // brute force:

 forall( current_edge, old_edges )
  if ( geom_segment_intersects_polygon( segment(
                                vg.inf( vg.source( current_edge ) ),
                                vg.inf( vg.target( current_edge ) ) ), pol ) )
   vg.del_edge( current_edge );

 vg_polygons.append( pol );
 return true;
}

bool vg_free( node current_node )
{
 polygon pol;
 forall( pol, vg_polygons )
  if ( pol.inside( vg.inf( current_node ) ) )
   return false;
 return true;
}

bool vg_free( node source_node, node target_node )
{
 polygon pol;
 segment s( vg.inf( source_node ), vg.inf( target_node ) );
 forall( pol, vg_polygons )
  if ( geom_segment_intersects_polygon( s, pol ) )
   return false;
 return true;
}

bool vg_belongs_to_polygon( node current_node , polygon pol )
{
  polygon cmp_pol = vg_node_polygon[current_node];
  return  pol.identical( cmp_pol );
}

node vg_insert_node( point p )
{
 polygon empty_pol;
 node insert_node = vg.new_node( p );
 vg_node_polygon[insert_node] = empty_pol;
 vg_wedge[insert_node] = vector( -PI, PI );
 return insert_node;
}

edge vg_insert_edge( node source_node, node target_node, bool mark )
{

// to be completed
 return vg.new_edge( source_node, target_node, mark );
}

bool vg_in_wedge( node source_node, node target_node )
{
 double angle = segment( vg.inf( source_node ),
                         vg.inf( target_node ) ).angle();
 vector wedge = vg_wedge[source_node];
 return ( ( ( wedge[0] <= wedge[1] ) && ( wedge[0] <= angle ) &&
            ( angle <= wedge[1] ) ) ||
          ( ( wedge[0] > wedge[1] ) && ( ( wedge[0] <= angle ) ||
                                         ( angle <= wedge[1] ) ) ) );
}

list<node> vg_rotational_sweep( node current_node, list<node>& old_nodes )
{
 point current_point;
 point old_point;
 node old_node;
 list<node> result;

/*

 seq_item current_seq_item;
 sortseq<double, node> sorted_old_nodes;

 forall( old_node, old_nodes )
  if ( vg_in_wedge( current_node, old_node ) &&
       vg_in_wedge( old_node, current_node ) )
  {
   current_point = vg.inf( current_node );
   old_point = vg.inf( old_node );
   angle = segment( current_point, old_point ).angle();
   if ( !( current_seq_item = sorted_old_nodes.lookup( angle ) ) ||
        ( current_point.distance( old_point ) <
          current_point.distance( vg.inf( sorted_old_nodes.inf(
                                   current_seq_item ) ) ) ) )
    sorted_old_nodes.insert( angle, old_node );
  }

*/

// brute force:

 polygon pol;
 bool intersects;
 segment s;

 forall( old_node, old_nodes )
 {
  if ( vg_in_wedge( current_node, old_node ) &&
       vg_in_wedge( old_node, current_node ) )
  {
   intersects = false;
   s = segment( vg.inf( current_node ), vg.inf( old_node ) );
   forall( pol, vg_polygons )
    if ( geom_segment_intersects_polygon( s, pol ) &&
         !( vg_belongs_to_polygon( current_node, pol ) ||
            vg_belongs_to_polygon( old_node, pol ) ) )
     intersects = true;
   if ( !intersects )
    result.append( old_node );
  }
 }

// end of brute force

 return result;
}

bool vg_tangent( node source_node, node target_node )
{
 bool source_tangent;
 bool target_tangent;
 polygon pol;
 segment s( vg.inf( source_node ), vg.inf( target_node ) );

 if ( source_node == start_node || source_node == goal_node )
  source_tangent = true;
 else
 {
  pol = vg_node_polygon[source_node];
  source_tangent = vg_tangent( s, pol );
 }

 if ( target_node == start_node || target_node == goal_node )
  target_tangent = true;
 else
 {
  pol = vg_node_polygon[target_node];
  target_tangent = vg_tangent( s, pol );
 }

 return ( source_tangent && target_tangent );
}

bool vg_tangent( segment s, polygon pol )
{
 int min_orientation = 0;
 int max_orientation = 0;
 int current_orientation;
 point p;
 forall( p, pol.vertices() )
 {
  current_orientation = orientation( p, s.source(), s.target() );
  min_orientation = Min( min_orientation, current_orientation );
  max_orientation = Max( max_orientation, current_orientation );
 }
 return ( min_orientation * max_orientation == 0 );
}

double vg_node_distance( node source_node, node target_node )
{
 return vg.inf( source_node ).distance( vg.inf( target_node ) );
}

list<segment> vg_search()
{
 if ( output_text )
  cout << "searching the vg (a*)..." << endl;

 list<segment> result;
 double inf = 1000 * geom_diameter();
 double cost;
 edge current_edge;
 node current_node;
 node neighbor = start_node;
 node_array<node> pred_nodes( vg );
 node_array<double> dist( vg );
 node_array<double> estimation( vg );
 node_pq<double> prio_queue( vg );

 forall_nodes( current_node, vg )
 {
  pred_nodes[current_node] = current_node;
  dist[current_node] = inf;
  estimation[current_node] = vg_node_distance( goal_node, current_node );
  prio_queue.insert( current_node, inf );
 }

 dist[start_node] = 0;
 prio_queue.decrease_p( start_node, estimation[start_node] );
 while ( (pred_nodes[goal_node] == goal_node ) &&
	 !prio_queue.empty())
 {
  current_node = prio_queue.del_min();

  forall_adj_edges( current_edge, current_node )
   if ( vg.inf( current_edge ) )
   {
    neighbor = vg.opposite( current_node, current_edge );
    cost = dist[current_node] + vg_node_distance( current_node, neighbor );
    if ( cost < dist[neighbor] )
    {
     dist[neighbor] = cost;
     pred_nodes[neighbor] = current_node;
     prio_queue.decrease_p( neighbor, ( cost + estimation[neighbor] ) );
    }
   }
 }

 if ( !(pred_nodes[goal_node] == goal_node ) )
 {
  current_node = goal_node;
  while( current_node != start_node )
  {
   result.push( vg_segment( pred_nodes[current_node], current_node ) );
   current_node = pred_nodes[current_node];
  }
 if ( output_text )
  cout << "...done. path of length " << dist[goal_node] << " found."
       << endl << endl;
 mp_dist = dist[goal_node];
 }
 else
 if ( output_text )
  cout << "...done. no path found." << endl << endl;

 return result;
}

void vg_draw_path( list<segment>& path )
{
 line_style old_line_style = win.set_line_style( path_style );
 vg_draw_path( path, path_color );
 win.set_line_style( old_line_style );
}

void vg_remove_path( list<segment>& path )
{
 vg_draw_path( path, BG_color );
 win_draw_polygons( vg_polygons, marked_color, marked_filled, false );
}

void vg_draw_path( list<segment>& path, color p_color )
{
 segment s;
 if ( !path.empty() )
  forall( s, path )
   win.draw_segment( s, p_color );
}

segment vg_segment( edge e )
{
 return segment( vg.inf( vg.source( e ) ), vg.inf( vg.target( e ) ) );
}

segment vg_segment( node source_node, node target_node )
{
 return segment( vg.inf( source_node ), vg.inf( target_node ) );
}

void mp_test( query_type qt, int min_start_goal_distance,
              int max_start_goal_distance, int number_of_passes )
{
 double min_dist = ( min_start_goal_distance / 100.0 ) * geom_diameter();
 double max_dist = ( max_start_goal_distance / 100.0 ) * geom_diameter();
 for ( int i = 1; i <= number_of_passes; i++ )
 {
  if ( output_text )
   cout << "test_series: no. " << i << " of " << number_of_passes << "."
        << endl << endl;
  do
  {
   start = random_point( win.xmin(), win.xmax(), win.ymin(), win.ymax() );
   goal = random_point( start, min_dist, max_dist );
  }
  while( ( goal.xcoord() < win.xmin() ) || ( goal.xcoord() > win.xmax() ) ||
         ( goal.ycoord() < win.ymin() ) || ( goal.ycoord() > win.ymax() ) ||
         geom_point_in_poly_list( start, poly_list ) ||
         geom_point_in_poly_list( goal, poly_list ) );
  start_goal_is_fixed = true;
  mp_start( qt, output_test_series );
 }
 if ( output_text )
  cout << "end of test series." << endl << endl;
}

void mp_start( query_type qt, int output )
{
 if ( objects_have_changed )
  mbt_construction();
 while ( !start_goal_is_fixed )
  win_menu_query_start_goal();

 if ( qt == em || qt == both )
  mp_ellipse_method( output );
 if ( qt == sm || qt == both )
  mp_segment_method( output );
}

void mp_ellipse_method( int output )
{
 if ( output_text )
  cout << "ENTERING ELLIPSE-METHOD" << endl << endl;
 mp_path.clear();
 mbt_set_all_marks( false );
 if ( output > 0 )
  win_draw_polygons( true );
 vg_init();

 bool mp_no_path = false;
 bool mp_ready = false;
 list<segment> path;
 path.append( segment( start, goal ) );
 mp_path = path;

 ellipse e;
 e.focus1 = start;
 e.focus2 = goal;
 e.diameter = start.distance( goal );

 while ( !( mp_no_path || mp_ready ) )
 {
  mp_objects = mbt_ellipse_query( e, false, true, true );
  if ( output > 1 )
  {
   win_draw_polygons( mp_objects, misc_no_path_color(), true, false );
   vg_draw_path( path );
  }
  if ( output > 2 )
   mp_export_query();
  if ( vg_update( mp_objects ) )
  {
   if ( output > 1 )
    vg_remove_path( path );
   path = vg_search();
   mp_path = path;
   if ( path.empty() )
    mp_no_path = true;
   else
    if ( mp_dist <= e.diameter || geom_path_in_ellipse( path, e ) )
     mp_ready = true;
    else
     e.diameter = Max( e.diameter * ellipse_factor, mp_dist );
  }
  else
  {
   mp_no_path = true;
   if ( output > 1 )
   {
    vg_remove_path( path );
      win_draw_polygons( false );
   }
  }
 }
 if ( mp_ready && output > 0 )
  vg_draw_path( path );

 mp_update_statistics( em, !mp_no_path, mp_dist, vg_polygons.size() );
 mp_objects.clear();
}

void mp_segment_method( int output )
{
 if ( output_text )
  cout << "ENTERING SEGMENT-METHOD" << endl << endl;
 mp_path.clear();
 mbt_set_all_marks( false );
 if ( output > 0 )
  win_draw_polygons( true );

 bool mp_no_path = false;
 bool mp_no_obstacles = false;
 segment s = vg_segment( vg_init() );
 list<segment> path;
 path.append( s );
 mp_path = path;

 while ( !( mp_no_path || mp_no_obstacles ) )
 {
  mp_objects.clear();
  forall( s, path )
   if ( mp_objects.empty() || sm_insert == 1 )
    mp_objects.conc( mbt_segment_query( s, false, true, true ) );
  if ( mp_objects.empty() )
  {
   mp_no_obstacles = true;
   if ( output > 0 )
    vg_draw_path( path );
  }
  else
  {
   if ( output > 1 )
   {
    win_draw_polygons( mp_objects, misc_no_path_color(), true, false );
    vg_draw_path( path );
   }
   if ( vg_update( mp_objects ) )
   {
    if ( output > 2 )
     mp_export_query();
    if ( output > 1 )
     vg_remove_path( path );
    path = vg_search();
    mp_path = path;
    if ( path.empty() )
     mp_no_path = true;
   }
   else
   {
    mp_no_path = true;
    if ( output > 1 )
    {
     vg_remove_path( path );
     win_draw_polygons( false );
    }
   }
  }
 }

 mp_update_statistics( sm, !mp_no_path, mp_dist, vg_polygons.size() );
 mp_objects.clear();
}

void mp_show_statistics()
{

}

void mp_update_statistics( query_type qt, bool success,
                           double abs_length,  int abs_number )
{
 double rel_length = 0;
 double rel_number = 0;
 if ( output_text )
 {
  if ( qt == em )
   cout << "ELLIPSE METHOD RESULT:" << endl;
  if ( qt == sm )
   cout << "SEGMENT METHOD RESULT:" << endl;
  if ( success )
  {
   rel_length = ( abs_length * 100.0 / geom_diameter() );
   rel_number = ( abs_number * 100.0 / poly_list.size() );
   cout << "length of the shortest path: " << abs_length << " ( "
        << rel_length << " % of diameter)" << endl;
   cout << "referred objects: " << abs_number << " ( "
        << rel_number << " % of all objects)" << endl << endl;
  }
  else
   cout << "there is no path connecting start and goal." << endl << endl;
 }

 mp_statistic_read();

 int i = ( int ) ( rel_length / 5 );
 if ( i < 0 )
  i = 0;
 if ( i > 40 )
  i = 40;

 if ( qt == em )
 {
  em_total_number++;
  if ( success )
  {
   em_success_number++;
   em_object_rate[i] =  em_object_rate[i] * em_object_count[i] + rel_number;
   em_object_count[i] = em_object_count[i] + 1;
   em_object_rate[i] = em_object_rate[i] / em_object_count[i];
  }
 }

 if ( qt == sm )
 {
  sm_total_number++;
  if ( success )
  {
   sm_success_number++;
   sm_object_rate[i] =  sm_object_rate[i] * sm_object_count[i] + rel_number;
   sm_object_count[i] = sm_object_count[i] + 1;
   sm_object_rate[i] = sm_object_rate[i] / sm_object_count[i];
  }
 }

 mp_statistic_write();
}

void mp_statistic_read()
{
 mp_statistic_reset();
 array<string> input( 396 );
 ifstream in( file_name + ".statistics" );
 int count = -1;
 while ( ++count < 396 )
  in >> input[count];
 em_total_number = atoi( input[45] );
 em_success_number = atoi( input[50] );
 sm_total_number = atoi( input[226] );
 sm_success_number = atoi( input[231] );
 for ( count = 0; count <= 40; count++ )
 {
  em_object_rate[count] = atof( input[4*count + 53] );
  em_object_count[count] = misc_atoi( input[4*count + 54] );
  sm_object_rate[count] = atof( input[4*count + 234] );
  sm_object_count[count] = misc_atoi( input[4*count + 235] );
 }
}

void mp_statistic_write()
{
 int i = 0;
 ofstream out( file_name + ".statistics" );
 out << flush;
 out << endl;
 out << "**************************" << endl;
 out << "* STATISTICAL EVALUATION *" << endl;
 out << "**************************" << endl << endl;
 out << "average number of referred objects ( % of all objects )" << endl;
 out << "in relation to path length ( % of diameter )" << endl;
 out << "grouped by path length intervals of 5%" << endl << endl << endl;

 out << "******************" << endl;
 out << "* ELLIPSE METHOD *" << endl;
 out << "******************" << endl << endl;
 out << "total number of queries     : " << em_total_number << endl;
 out << "number of successful queries: " << em_success_number << endl << endl;
 for ( i = 0; i < 40; i++ )
  out << i * 5 << "%- " << ( i + 1 ) * 5 << "%:  " << em_object_rate[i]
      << "  (" << em_object_count[i] << ")" << endl;
 out << "    > 200%:  " << em_object_rate[40]
     << "  (" << em_object_count[40] << ")" << endl << endl << endl;

 out << "******************" << endl;
 out << "* SEGMENT METHOD *" << endl;
 out << "******************" << endl << endl;
 out << "total number of queries     : " << sm_total_number << endl;
 out << "number of successful queries: " << sm_success_number << endl << endl;
 for ( i = 0; i < 40; i++ )
  out << i * 5 << "%- " << ( i + 1 ) * 5 << "%:  " << sm_object_rate[i]
      << "  (" << sm_object_count[i] << ")" << endl;
 out << "    > 200%:  " << sm_object_rate[40]
     << "  (" << sm_object_count[40] << ")" << endl << endl << endl;
}

void mp_statistic_reset()
{
 array<int> empty_object_count(41);
 array<double> empty_object_rate(41);

 em_total_number = 0;
 em_success_number = 0;
 em_object_count = empty_object_count;
 em_object_rate = empty_object_rate;
 sm_total_number = 0;
 sm_success_number = 0;
 sm_object_count = empty_object_count;
 sm_object_rate = empty_object_rate;
}

void mp_export_query()
{
 double x_input, y_input;
 window *window_pointer;

 int mouse_button = read_mouse( window_pointer, x_input, y_input );
 if ( ( window_pointer == &win ) && ( mouse_button == 4 ) )
  misc_export_to_eps();
}

polygon random_polygon( int max_number_of_segments, double max_size )
{
 list<point> point_list;
 point anchor = random_point( win.xmin(), win.xmax(), win.ymin(), win.ymax() );
 point_list.append( anchor );
 int number_of_segments = random( 3, max_number_of_segments );
 for ( int i = 2; i <= number_of_segments; i++ )
  point_list.append( random_point( anchor, 0, max_size/2 ) );
 return CONVEX_HULL( point_list );
}

point random_point( point anchor, double min_size, double max_size )
{
 double angle;
 double distance;
 random >> angle;
 random >> distance;
 angle =  PI * ( 2 * angle - 1 );
 distance = distance * ( max_size - min_size ) + min_size;
 return anchor.translate( angle, distance );
}

point random_point( double xmin, double xmax, double ymin, double ymax )
{
 double x;
 double y;
 random >> x;
 random >> y;
 return point( xmin + x * ( xmax - xmin ), ymin + y * ( ymax - ymin ) );
}

color misc_new_color()
{
 if ( marked_color != color( red ) && other_color != color( red ) )
  return color( red );
 else
  if ( marked_color != color( yellow ) && other_color != color( yellow ) )
   return color( yellow );
  else
   return color( orange );
}

color misc_another_new_color()
{
 if ( marked_color != color( green ) && other_color != color( green ) )
  return color( green );
 else
  if ( marked_color != color( blue ) && other_color != color( blue ) )
   return color( blue );
  else
   return color( violet );
}

color misc_no_path_color()
{
 if ( marked_color != color( yellow ) && other_color != color( yellow )
                             && path_color != color( yellow ) )
  return color( yellow );
 else
  if ( marked_color != color( orange ) && other_color != color( orange )
                              && path_color != color( orange ) )
   return color( orange );
  else
   if ( marked_color != color( red ) && other_color != color( red )
                                     && path_color != color( red ) )
    return color( red );
   else
    return color( pink );
}

int misc_atoi( string s )
{
 s = s.replace_all( "(", "" );
 s = s.replace_all( ")", "" );
 return atoi( s );
}

void misc_export_to_eps()
{
 list<point> start_goal_list;
 if ( start_goal_is_fixed )
 {
  start_goal_list.append( start );
  start_goal_list.append( goal );
 }
 else
  mp_path.clear();
 string film_name( "-%d.eps", film_index );
 film_name = file_name + film_name;
 string old_name = film_name;
 list<polygon> pl1;
 list<polygon> pl2;
 list<polygon> pl3;

 panel export_panel;
 export_panel.string_item( "export to file: ", film_name );
 export_panel.button( "OK" );
 export_panel.button( "CANCEL" );
 export_panel.buttons_per_line( 2 );

 int button_number = export_panel.open( win );
 if ( button_number == 0 )
 {
  if ( film_name == old_name )
   film_index++;
  if ( output_text )
   cout << "export to " << film_name << endl << endl;
  if ( objects_have_changed )
   pl1 = poly_list;
  else
  {
   pl1 = mbt_get_tree( false, true, false );
   pl2 = mbt_get_tree( true, false, false );
  }
  pl3 = mp_objects;

  convert_to_postscript( start_goal_list, mp_path, pl1, pl2, pl3,
                         export_fill1, export_fill2, export_fill3,
                         film_name, win.xmin(), win.xmax(),
                         win.ymin(), win.ymax() );
 }
}

/* Die Objekte sollen in einem File mit dem Name file_name
gespeichert werden. Die Polygone sollen OHNE die berandenden
Strecken gezeichnet werden, gefuellt mit Fuellgrad fill1
fuer poly_list1 bzw. fill2 fuer poly_list2. Dabei ist
0 <= fill1, fill2 <= 100. */

void convert_to_postscript( list<point>& point_list,
                            list<segment>& segment_list,
                            list<polygon>& poly_list1,
                            list<polygon>& poly_list2,
                            list<polygon>& poly_list3,
                            int fill1, int fill2, int fill3,
                            string file_name,
			    double min_x, double max_x,
		            double min_y, double max_y)
{
  point punkt;
  segment strecke;
  polygon pol;

  // Name der Ausgabe(-ps)-Datei
  ofstream out(file_name);

  // Leere die Ausgabedatei
  out << flush;
  out << "%!PS-Adobe-2.0 EPSF-2.0\n";
  out << "%%Title: Paper-Zuarbeit\n";
  out << "%%Creator: C++\n";
  out << "%%CreationDate: " << time((time_t *)0) << endl;
  out << "%%For: LaTex\n";
  out << "%%Pages: 1\n";
  out << "%%DocumentFonts:\n";        // Fonts belanglos fuer diese Anwendung
  out << "%%BoundingBox: " << min_x << ' ' << min_y << ' '
                           << max_x << ' ' << max_y << '\n';
  out << "%%EndComments\n";
  out << "%%EndProlog\n\n";

  out << "0 0 0 setrgbcolor\n";
  out << "0 setlinecap\n";
  out << "0 setlinejoin\n";
  out << "0.5 setlinewidth\n\n";

  // Zeichne Bounding Box
  out << "% Box:\n";
  out << "0 setgray\n";
  out << "newpath\n";
  out <<  min_x << ' ' <<  min_y << " moveto\n";
  out <<  min_x << ' ' <<  max_y << " lineto\n";
  out << "fill\n";
  out <<  min_x << ' ' <<  max_y << " moveto\n";
  out <<  max_x << ' ' <<  max_y << " lineto\n";
  out << "fill\n";
  out <<  max_x << ' ' <<  max_y << " moveto\n";
  out <<  max_x << ' ' <<  min_y << " lineto\n";
  out << "fill\n";
  out <<  max_x << ' ' <<  min_y << " moveto\n";
  out <<  min_x << ' ' <<  min_y << " lineto\n";
  out << "fill\n\n";

  // Zeichne Punkte als kleine Scheibe.
  if (!point_list.empty())
    forall(punkt,point_list) {
      out << "% (Dicker) Punkt:\n";
      out << "newpath\n";
      out << punkt.xcoord() << ' ' << punkt.ycoord()
	  << ' ' << ((max_x - min_x)/300) << " 0 360 arc\n";
      out << "fill\n\n"; }

  // Zeichne Polygone mit angegebener Fuelldichte fill1.
  if (!poly_list1.empty())
    forall(pol,poly_list1) {
      out << "% Polygon:\n";
      out << "newpath\n";
      out << pol.vertices().head().xcoord() << ' '
          << pol.vertices().head().ycoord() << " moveto\n";
      forall(punkt,pol.vertices())
        if (pol.vertices().rank(punkt)!=1)
	   out << punkt.xcoord() << ' '
               << punkt.ycoord() << " lineto\n";
      out << (1-fill1/100.0) << " setgray\n";
      out << "fill\n\n";
   }

  // Zeichne Polygone mit angegebener Fuelldichte fill2.
  if (!poly_list2.empty())
    forall(pol,poly_list2) {
      out << "% Polygon:\n";
      out << "newpath\n";
      out << pol.vertices().head().xcoord() << ' '
          << pol.vertices().head().ycoord() << " moveto\n";
      forall(punkt,pol.vertices())
        if (pol.vertices().rank(punkt)!=1)
	   out << punkt.xcoord() << ' '
               << punkt.ycoord() << " lineto\n";
      out << (1-fill2/100.0) << " setgray\n";
      out << "fill\n\n"; }

  // Zeichne Polygone mit angegebener Fuelldichte fill3.
  if (!poly_list3.empty())
    forall(pol,poly_list3) {
      out << "% Polygon:\n";
      out << "newpath\n";
      out << pol.vertices().head().xcoord() << ' '
          << pol.vertices().head().ycoord() << " moveto\n";
      forall(punkt,pol.vertices())
        if (pol.vertices().rank(punkt)!=1)
	   out << punkt.xcoord() << ' '
               << punkt.ycoord() << " lineto\n";
      out << (1-fill3/100.0) << " setgray\n";
      out << "fill\n\n"; }

  // Zeichne Segmente. Sie werden durch "fill" und nicht durch "stroke"
  // sichtbar gemacht aus Implementationsgruenden.
  // (Liniendicke unabhaengig von der Skalierung)
  if (!segment_list.empty())
    forall(strecke,segment_list) {
      out << "% Strecke:\n";
      out << "newpath\n";
      out << strecke.start().xcoord() << ' '
          << strecke.start().ycoord() << " moveto\n";
      out << strecke.end  ().xcoord() << ' '
          << strecke.end  ().ycoord() << " lineto\n";
      out << "0.0 setgray\n";
      out << "fill\n\n";
   }
}

